import os
from contextlib import contextmanager
from typing import Optional

import polars as pl
import snowflake.connector as snow
from snowflake.connector.pandas_tools import write_pandas

from klondike import logger
from klondike.utilities.utilities import validate_if_exists_behavior

##########


class SnowflakeConnector:
    """
    Leverages connection to Snowflake to read and write Polars DataFrame
    objects to the data warehouse

    `Args`:
        snowflake_user: `str`
            Username to connect to Snowflake (defaults to `SNOWFLAKE_USER` in environment)
        snowflake_password: `str`
            Password to connect to Snowflake (defaults to `SNOWFLAKE_PASSWORD` in environment)
        snowflake_account: `str`
            Account identifier for Snowflake warehouse (defaults to `SNOWFLAKE_ACCOUNT` in environment)
        snowflake_warehouse: `str`
            Snowflake warehouse name (defaults to `SNOWFLAKE_WAREHOUSE` in environment)
        snowflake_database: `str`
            Snowflake database name (defaults to `SNOWFLAKE_DATABASE` in environment)
        row_chunk_size: `int`
            Default row chunk size for reading from / writing to Snowflake
    """

    def __init__(
        self,
        snowflake_user: Optional[str] = None,
        snowflake_password: Optional[str] = None,
        snowflake_account: Optional[str] = None,
        snowflake_database: Optional[str] = None,
        snowflake_warehouse: Optional[str] = None,
        row_chunk_size: int = 100_000,
    ):
        """
        All Snowflake connection values either need to be supplied as constructor
        arguments or inferred from the environment; if neither occurs, a `ValueError`
        will be raised
        """

        self.snowflake_user = (
            snowflake_user if snowflake_user else os.getenv("SNOWFLAKE_USER")
        )
        self.snowflake_password = (
            snowflake_password
            if snowflake_password
            else os.getenv("SNOWFLAKE_PASSWORD")
        )
        self.snowflake_account = (
            snowflake_account if snowflake_account else os.getenv("SNOWFLAKE_ACCOUNT")
        )
        self.__snowflake_warehouse = (
            snowflake_warehouse
            if snowflake_warehouse
            else os.getenv("SNOWFLAKE_WAREHOUSE")
        )
        self.__snowflake_database = (
            snowflake_database
            if snowflake_database
            else os.getenv("SNOWFLAKE_DATABASE")
        )

        ###

        self.__validate_authentication()

        ###

        self.dialect = "snowflake"
        self.__row_chunk_size = row_chunk_size

    def __validate_authentication(self):
        _auth_vals = [
            self.snowflake_user,
            self.snowflake_password,
            self.snowflake_account,
            self.snowflake_database,
            self.snowflake_warehouse,
        ]

        if any([not x for x in _auth_vals]):
            raise ValueError(
                "Missing authentication values! Make sure all `snowflake_*` values are provided at construction"
            )

    @property
    def snowflake_warehouse(self):
        return self.__snowflake_warehouse

    @snowflake_warehouse.setter
    def snowflake_warehouse(self, warehouse):
        self.__snowflake_warehouse = warehouse

    @property
    def snowflake_database(self):
        return self.__snowflake_database

    @snowflake_database.setter
    def snowflake_database(self, database):
        self.__snowflake_database = database

    @property
    def row_chunk_size(self):
        return self.__row_chunk_size

    @row_chunk_size.setter
    def row_chunk_size(self, row_chunk_size):
        self.__row_chunk_size = row_chunk_size

    @contextmanager
    def connection(self):
        """
        Creates a connection to Snowflake
        """

        conn = snow.connect(
            account=self.snowflake_account,
            warehouse=self.snowflake_warehouse,
            user=self.snowflake_user,
            password=self.snowflake_password,
        )

        try:
            yield conn
            conn.commit()
        finally:
            conn.close()

    @contextmanager
    def cursor(self, connection):
        """
        Leverages Snowflake connection to execute SQL transactions
        """

        cur = connection.cursor()

        try:
            yield cur
        finally:
            cur.close()

    def __query(self, sql: str):
        """
        Executes SQL command against Snowflake warehouse

        Args:
            sql: `str`
                SQL query in string format
        """

        with self.connection() as _conn:
            with self.cursor(_conn) as _cursor:
                # TODO - Ugly! Clean this up
                _cursor.execute(f"USE DATABASE {self.snowflake_database};")
                _cursor.execute(f"USE WAREHOUSE {self.snowflake_warehouse};")
                _cursor.execute(sql)
                _resp = _cursor.fetch_arrow_batches()

        try:
            return pl.from_arrow(_resp)
        except ValueError as ve:
            # NOTE - This appears to be Polars response to empty fetch_arrow_batches()
            # This should be interrogated more, but is functional
            if "Must pass schema, or at least one RecordBatch" in str(ve):
                logger.debug("No results obtained via query")
                return pl.DataFrame()
            raise

    def read_dataframe(self, sql: str) -> pl.DataFrame:
        """
        Executes a SQL query against the Snowflake warehouse
        and returns the result as a Polars DataFrame object

        Args:
            sql: `str`
                String representation of SQL query

        Returns:
            Polars DataFrame object
        """

        # Execute SQL against warehouse
        logger.debug("Running SQL...", sql)
        df = self.__query(sql=sql)

        logger.info(f"Successfully read {len(df)} rows from Snowflake")

        return df

    def write_dataframe(
        self,
        df: pl.DataFrame,
        table_name: str,
        schema_name: str,
        database_name: Optional[str] = None,
        if_exists: str = "append",
        auto_create_table: bool = True,
        chunk_output: bool = False,
    ) -> None:
        """
        Writes a Polars DataFrame to Snowflake

        TODO - Utilizing `.to_pandas()` may not be ideal, we should consider
        modifying this function to utilize `PyArrow` under the hood (see - `klondike.BigQueryConnector`)

        Args:
            df: `pl.DataFrame`
                Polars DataFrame to be written to Snowflake
            table_name: `str`
                Destination table name
            database_name: `str`
                Destination database name
            schema_name: `str`
                Destination schema name
            database: `str`
                Optional Snowflake database (defaults to preset class attribute)
            if_exists: `str`
                One of `append`, `truncate`, `drop`, `fail`
            auto_create_table: `bool`
                If true, the desintation table will be created if it doesn't already exist
            chunk_output: `int`
                If true, the default chunk size will be applied; otherwise no chunking will occur
        """

        # Use class attribute  for row chunking
        row_chunks = self.row_chunk_size if chunk_output else None

        # Confirm that user has passed in valid logic
        if not validate_if_exists_behavior(if_exists):
            raise ValueError(f"{if_exists} is an invalid input")

        # This logic leverages the behavior of the write_pandas() function ... see below
        if if_exists == "drop" and auto_create_table:
            overwrite = True
            logger.warning(f"{table_name} will be dropped if it exists")

        elif if_exists == "truncate":
            overwrite = True
            auto_create_table = False

        elif if_exists == "append":
            overwrite = False
            auto_create_table = False

        elif if_exists == "fail":
            if self.table_exists(schema_name=schema_name, table_name=table_name):
                raise snow.errors.DatabaseError(f"{table_name} already exists")

        database = database_name if database_name else self.snowflake_database

        ###

        logger.info(
            f"Writing to {self.snowflake_database}.{schema_name}.{table_name}..."
        )
        with self.connection() as conn:
            resp, num_chunks, num_rows, output = write_pandas(
                conn=conn,
                df=df.to_pandas(),
                database=database,
                schema=schema_name,
                table_name=table_name,
                auto_create_table=auto_create_table,
                chunk_size=row_chunks,
                overwrite=overwrite,
            )

        ###

        if resp:
            logger.info(f"Successfully wrote {num_rows} rows to {table_name}")
        else:
            logger.error(f"Failed to write to {table_name}", resp)
            raise

    def table_exists(
        self, table_name: str, database_name: Optional[str] = None
    ) -> bool:
        """
        Determines if a Snowflake table exists in the warehouse

        Args:
            table_name: `str`
                Snowflake table name in `schema.table` format

        Returns:
            True if the table exists, False otherwise
        """

        schema, table = table_name.split(".")

        if not database_name:
            logger.debug(f"Defaulting to default database [{self.snowflake_database}]")
            database_name = self.snowflake_database

        sql = f"""
        SELECT
            *
        FROM INFORMATION_SCHEMA.TABLES
        WHERE table_schema = '{schema}'
            AND table_name = '{table}'
            AND table_catalog = '{database_name}'
        """

        resp = self.__query(sql=sql)

        return not resp.is_empty()

    def list_tables(
        self, schema_name: str, database_name: Optional[str] = None
    ) -> list:
        """
        Gets a list of available tables in a Snowflake schema

        Args:
            schema_name: `str`
            database_name: `str`

        Returns:
            List of table names
        """

        if not database_name:
            database_name = self.snowflake_database

        sql = f"""
        SELECT
            table_catalog,
            table_schema,
            table_name
        FROM INFORMATION_SCHEMA.TABLES
        WHERE table_schema = '{schema_name}'
            AND table_catalog = '{database_name}'
        """

        resp = self.__query(sql=sql)

        resp = resp.select(table_name=pl.concat_str(resp.columns, separator="."))

        return pl.Series(resp["table_name"]).to_list()



"""
General utilites to use across all database connections
"""


def validate_if_exists_behavior(
    user_input: str, acceptable_values: list = ["fail", "append", "truncate", "drop"]
):
    """
    Ensures that user input to `write_dataframe` function is valid
    """

    return user_input in acceptable_values



import io
import json
import os
import tempfile
from typing import Optional, Union

import polars as pl
from google.cloud import bigquery
from google.cloud.bigquery import LoadJobConfig
from google.cloud.exceptions import NotFound

from klondike import logger
from klondike.utilities.utilities import validate_if_exists_behavior

##########

SCOPES = (
    {
        "scopes": [
            "https://www.googleapis.com/auth/bigquery",
        ]
    },
)[0]


class BigQueryConnector:
    """
    Establish and authenticate a connection to a BigQuery warehouse

    Args:
        app_creds: `str`
            Google service account, either as a relative path or a dictionary instance
        project: `str`
            Name of Google Project
        location: `str`
            Location of Google Project
        timeout: `int`
            Temporal threshold to kill a stalled job, defaults to 60s
        client_options: `list`
            API scopes
        google_environment_variable: `str`
            Provided for flexibility, defaults to `GOOGLE_APPLICATION_CREDENTIALS`
    """

    def __init__(
        self,
        app_creds: Optional[Union[str, dict]] = None,
        project: Optional[str] = None,
        location: Optional[str] = None,
        timeout: int = 60,
        client_options: dict = SCOPES,
        google_environment_variable: str = "GOOGLE_APPLICATION_CREDENTIALS",
    ):
        self.app_creds = app_creds
        self.project = project
        self.location = location
        self.client_options = client_options

        self.dialect = "bigquery"
        self.__client = None
        self.__timeout = timeout

        if not self.app_creds:
            if not os.environ.get(google_environment_variable):
                raise OSError("No app_creds provided")
            else:
                logger.info(
                    f"Using `{google_environment_variable}` variable defined in environment"
                )
        else:
            self.__setup_google_app_creds(
                app_creds=self.app_creds, env_variable=google_environment_variable
            )

    @property
    def client(self):
        """
        Instantiate BigQuery client and assign it
        as class property
        """

        if not self.__client:
            self.__client = bigquery.Client(
                project=self.project,
                location=self.location,
                client_options=self.client_options,
            )

        return self.__client

    @property
    def timeout(self):
        return self.__timeout

    @timeout.setter
    def timeout(self, timeout):
        self.__timeout = timeout

    def __setup_google_app_creds(self, app_creds: Union[str, dict], env_variable: str):
        "Sets runtime environment variable for Google SDK"

        if isinstance(app_creds, dict):
            creds = json.dumps(app_creds)
            tempp_file = tempfile.TemporaryFile(suffix=".json")

            with open(tempp_file.name, "w") as f:
                f.write(creds)

        elif isinstance(app_creds, str):
            creds = app_creds

        os.environ[env_variable] = creds

    def __set_load_job_config(
        self,
        base_job_config: Optional[LoadJobConfig] = None,
        max_bad_records: int = 0,
        table_schema: list = None,
        if_exists: str = "fail",
        **kwargs,
    ):
        "Defines `LoadConfigJob` when writing to BigQuery"

        def set_write_disposition(if_exists: str):
            DISPOSITION_MAP = {
                "fail": bigquery.WriteDisposition.WRITE_EMPTY,
                "append": bigquery.WriteDisposition.WRITE_APPEND,
                "truncate": bigquery.WriteDisposition.WRITE_TRUNCATE,
            }

            return DISPOSITION_MAP[if_exists]

        def set_table_schema(table_schema: list):
            return [bigquery.SchemaField(**x) for x in table_schema]

        ###

        if not base_job_config:
            logger.debug("No job config provided, starting fresh")
            base_job_config = LoadJobConfig()

        # This is default behavior for this class
        base_job_config.source_format = bigquery.SourceFormat.PARQUET

        # Create table schema mapping if provided
        if table_schema:
            base_job_config.schema = set_table_schema(table_schema=table_schema)
        else:
            base_job_config.schema = None

        base_job_config.max_bad_records = max_bad_records
        base_job_config.write_disposition = set_write_disposition(if_exists=if_exists)

        ###

        # List of available LoadJobConfig attributes
        _attributes = [x for x in dict(vars(LoadJobConfig)).keys()]

        # Attributes that will not be overwritten
        _reserved_attributes = [
            "source_format",
            "schema",
            "max_bad_records",
            "write_disposition",
        ]

        # Loop through keyword arguments and update `LoadJobConfig` object
        for k, v in kwargs.items():
            if k in _attributes and k not in _reserved_attributes:
                logger.debug(f"Updating {k} parameter in job config [value={v}]")
                base_job_config[k] = v

        return base_job_config

    def read_dataframe(self, sql: str) -> pl.DataFrame:
        """
        Executes a SQL query and returns a Polars DataFrame.
        TODO - Make this more flexible and incorporate query params

        Args:
            sql: `str`
                String representation of SQL query

        Returns:
            Polars DataFrame object
        """

        # Define query job
        logger.debug("Running SQL...", sql)
        query_job = self.client.query(query=sql, timeout=self.timeout)

        # Execute and wait for results
        result = query_job.result()

        # Populate DataFrame using PyArrow
        df = pl.from_arrow(result.to_arrow())

        if df.is_empty():
            logger.info("No results returned from SQL call")
            return

        logger.info(f"Successfully read {len(df)} rows from BigQuery")

        return df

    def write_dataframe(
        self,
        df: pl.DataFrame,
        table_name: str,
        load_job_config: Optional[LoadJobConfig] = None,
        max_bad_records: int = 0,
        table_schema: list = None,
        if_exists: str = "fail",
        **load_kwargs,
    ) -> None:
        """
        Writes a Polars DataFrame to BigQuery

        Args:
            df: `polars.DataFrame`
                DataFrame to write to BigQuery
            table_name: `str`
                Destination table name to write to - `dataset.table` convention
            load_job_config: `LoadJobConfig`
                Configures load job; if none is supplied, several defaults are applied
            max_bad_records: `int`
                Tolerance for bad records in the load job, defaults to 0
            table_schema: `list`
                List of column names, types, and optional flags to include
            if_exists: `str`
                One of `fail`, `drop`, `append`, `truncate`
            load_kwargs:
                See here for list of accepted values \
                    https://cloud.google.com/python/docs/reference/bigquery/latest/google.cloud.bigquery.job.LoadJobConfig
        """

        if not validate_if_exists_behavior(user_input=if_exists):
            raise ValueError(f"{if_exists} is an invalid input")

        if if_exists == "drop":
            self.client.delete_table(table=table_name)
            if_exists = "fail"

        load_job_config = self.__set_load_job_config(
            base_load_conifg=load_job_config,
            max_bad_records=max_bad_records,
            table_schema=table_schema,
            if_exists=if_exists,
            **load_kwargs,
        )

        logger.info(f"Writing to {table_name}...")
        with io.BytesIO() as stream_:
            df.write_parquet(stream_)
            stream_.seek(0)
            load_job = self.client.load_table_from_file(
                stream_, destination=table_name, job_config=load_job_config
            )

        load_job.result()
        logger.info(f"Successfuly wrote {len(df)} rows to {table_name}")

    def table_exists(self, table_name: str) -> bool:
        """
        Determines if a BigQuery table exists

        Args:
            table_name: `str`
                BigQuery table name in `schema.table` or `project.schema.table` format
        """

        try:
            _ = self.client.get_table(table=table_name)
            return True

        except NotFound:
            return False

    def list_tables(self, schema_name: str) -> list:
        """
        Gets a list of available tables in a BigQuery schema

        Args:
            schema_name: `str`
                BigQuery schema name

        Returns:
            List of table names
        """

        return [
            x.full_table_id.replace(":", ".")
            for x in self.client.list_tables(dataset=schema_name)
        ]



import os
import pathlib

import setuptools

##########


def main():
    HERE = pathlib.Path(__file__).parent
    README = (HERE / "README.md").read_text()

    with open(os.path.join(HERE, "requirements.txt")) as requires_:
        INSTALL_REQUIRES = requires_.read().strip().split("\n")

    setuptools.setup(
        name="klondike",
        version="0.2.0",
        author="Ian Richard Ferguson",
        author_email="IRF229@nyu.edu",
        url="https://github.com/IanRFerguson/klondike",
        keywords=["API", "ETL", "BIGQUERY", "SNOWFLAKE"],
        packages=setuptools.find_packages(),
        install_requires=INSTALL_REQUIRES,
        classifiers=[
            "Intended Audience :: Developers",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
        ],
        python_requires=">=3.8.0,<3.11.0",
        long_description=README,
        long_description_content_type="text/markdown",
    )


#####

if __name__ == "__main__":
    main()



import importlib
import logging
import os

# Define project logger
logger = logging.getLogger(__name__)

_handler = logging.StreamHandler()
_formatter = logging.Formatter("%(module)s %(levelname)s %(message)s")
_handler.setFormatter(_formatter)

logger.addHandler(_handler)

# Define logging behavior
if any([os.environ.get("TESTING"), os.environ.get("TEST")]):
    logger.setLevel("WARNING")
elif os.environ.get("DEBUG"):
    logger.setLevel("DEBUG")
    logger.debug("Debugger active...")
else:
    logger.setLevel("INFO")

#####

__all__ = []
POLAR_OBJECTS = [
    ("klondike.bigquery.bigquery", "BigQueryConnector"),
    ("klondike.snowflake.snowflake", "SnowflakeConnector"),
]

for module_, object_ in POLAR_OBJECTS:
    try:
        globals()[object_] = getattr(importlib.import_module(module_), object_)
    except ImportError:
        logger.debug(f"Could not import {module_}.{object_} ... skipping")
